function challenge(recordName)
[~,config]=wfdbloadlib;
if(config.inOctave)
    crash_dumps_octave_core(0);
end
annName = 'qrs'; % This annotator name is required by the Challenge rules.
[siginfo, fs] = wfdbdesc2(recordName);
description = squeeze(struct2cell(siginfo));   
descriptionNamesTable = description(5,:);
if(~isnan(fs))
    Fs = fs(1);
else
    Fs=sscanf(siginfo(1).SamplingFrequency,'%f');
end

if(config.inOctave)
    stringInput =sprintf('wfdb2mat -r %s -H',recordName);
    system(stringInput);
    load(strcat(recordName, 'm.mat'));
    signal=val(1:end,:)';
    tm=(1:size(signal,1))./Fs;
    for ii=1:size(signal,2)
        gain=sscanf(siginfo(ii).Gain,'%f');
        base=sscanf(siginfo(ii).Baseline,'%f');
        signal(:,ii)=(signal(:,ii)-base)./gain;
    end
end
ecg = signal(:,1);
fc = [48 52]*2*(1/Fs);                                                   % set cut-off frequencies
[B,A] = butter(3,fc,'stop');                                        % construct bandpass filters
ecg = filtfilt(B,A,ecg);                                    % filter bp-signal
fc = [58 62]*2*(1/Fs);                                                   % set cut-off frequencies
[B,A] = butter(3,fc,'stop');                                        % construct bandpass filters
ecg = filtfilt(B,A,ecg);                                    % filter bp-signal
try % on some data-sets the gqrs-function yields no results
    gqrs(recordName)
    ann_ecg = rdann(recordName,annName);
catch
    D_s_filt = diff(ecg);                                                % derative of filtered signal
    D_s_filt(D_s_filt < 0) = 0;                                             % set all signal with non-positve slope to zero
    [ann_ecg] = FP(D_s_filt,round(0.375*Fs));      % identify prominent peaks as bp-pulses (min peak distance is at least .5 seconds)
end
ann_ecg = ann_ecg'; % transpose the annotation-array
try
[ann_bpa] = check_BP(signal,Fs); % this function generates abp annotations by checking all remaining channels for the most suitable pressure-signal
if isempty(ann_bpa) == 0 && numel(ann_bpa) > 2 % at least 2 abp-annotations have to be generated
    [ann_ecg,delay] = check_PM(recordName,ann_ecg,ann_bpa,Fs); % this function checks for the presence of a PM, and generates the ecg-annotations if a PM is identified, it also return the delay between the ABP- and ECG-annotations
end
new_ann_ecg = []; % pre-allocate a variable for possible additional ecg-annotations
ann_idx = 0; % an index for the annotations to add

% this loop checks for possible missing ecg-annotation by if pair of subsequent ABP-annotations sandwiches an ECG-annotations
for i = 1:length(ann_bpa)-1 % run the loop for the number of ABP-annotations minus 1 (eg. the number of pairs)
    a = ann_ecg(ann_ecg>ann_bpa(i)); % all ECG-annotations after the current ABP-annotations
    if isempty(a) == 0 % check whether there are actually any ECG-annotations after the current ABP-annotations
        if a(1) > ann_bpa(i+1) % if the first ECG-annotation following the current ABP-annotation is later than the first subsequent ABP-annotation an ECG-annotation is probably missing
            ann_idx = ann_idx + 1; % raise the index for the annotations to add
            new_ann_ecg(ann_idx) = ann_bpa(i+1)-delay; % generate a new ECG-annotations at the site of the first subsequent ABP-annotation minus the previously obtained delay
        end
    end
end


if iscolumn(ann_ecg) == 1 % make sure ann_ecg is a horizontal array
    ann_ecg = ann_ecg';
end

ann_ecg = [new_ann_ecg ann_ecg]; % add the new ecg-annotations to the old ecg-annotations

new_ann_ecg = []; % clear the array for the new annotations to add
a = ann_bpa(ann_bpa>ann_ecg(end)); % check whether there are any ABP-annotations after the last ECG-annotation
if numel(a) > 1 % if there are such annotations, add these annotations (corrected for the delay, to the ECG-annotations)
    new_ann_ecg = a-delay;
end
ann_ecg = [new_ann_ecg ann_ecg]; % add the new ecg-annotations to the old ecg-annotations

RA_idx = 0;
RA = []; % pre-allocate array of annotations which shall be removed
ann_ecg = sort(ann_ecg); % sort the ecg-annotations
for i = 1:length(ann_ecg) % run through all ecg-annotations
    % for the current ecg-annotations, check the location of the first subsequent ABP-annotation
    % thereafter check if there are any ECG-annotations between these two locations
    a = ann_bpa(ann_bpa>ann_ecg(i)); % all ABP-annotations after the current ECG-annotation
    if isempty(a) == 0 && a(1)-ann_ecg(i) < 2*Fs % if there are ABP-annotations after the current ECG-annotation they cannot be further away than 2 seconds
        ecg_idxs = ann_ecg(i+1:end); % all ecg-annotations after the current ecg-annotation
        ecg_idxs = ecg_idxs(ecg_idxs < a(1)); % all ecg-annotations after the current ecg-annotation and prior to the first subsequent ABP-annotations
        if isempty(ecg_idxs) == 0 % check whether there are such annotations
            ii = length(ecg_idxs); % the number of annotations after the current ecg-annotation and prior to the first subsequent ABP-annotation (potential false positives)
            while ii > 0 % run through all of these before mentioned annotions
                index = HLC(ecg_idxs(ii),ecg,0.2*Fs); % check for all of these annotations whether they are noisy
                if index == 1 % if annotation if noisy, add to array of annotations which shall be removed
                    RA_idx = RA_idx + 1; % raise the index
                    RA(RA_idx) = ecg_idxs(ii); % add annotations to array
                end
                ii = ii - 1;
            end
        end
    end
end
for i = 1:RA_idx
    ann_ecg(ann_ecg == RA(i)) = [];
end
if iscolumn(ann_ecg) == 0 % make sure ann-ecg is a column
    ann_ecg = ann_ecg';
end
end
wrann(recordName,'qrs',ann_ecg,[],[],[],[]); % write the ecg-annotations
end
function [ann_bpa] = check_BP(signal,Fs)
% check whether there are any pressure signals suitable for analysis, if none are suitable return empty array
FsBU = Fs; % Back-up the original sampling frequency
fact = round(Fs/80); % determine down-sampling factor (downsample to approximately 80 Hz)
signalBU = signal; % back-up the original signals
signal = downsample(signal,fact); % downsample the signal with previously determined factor
Fs = Fs/fact; % recalculate sampling frequency
[~,numsigs] = size(signal); % determine number of channels
for i = 1:numsigs-1 % run to all channels except the ecg-channel(thus the first one)
    clear rel_RR
    abp_cand{i} = ABP(signal(:,i+1),.375*Fs,0.0833*Fs); % generate candidate annotations for the current signal and the ABP-algorithm and a minimal peak-distance of 0.375 seconds
    RR = diff(abp_cand{i}); % the RR-intervals between the candiate annotations
    if isempty(RR) == 0 % check whether there are annotations generated
        [freq,power] = hist(RR,100); % construct 100 bin-histogram of intervals
        [~,idx] = max(freq); % determine most frequent interval
        MPI = power(idx); % the Most Prominent Interval
        rel_RR = zeros(1,length(RR)-1); % pre-allocate array
        for ii = 1:length(RR)-1 % run through all interval-pairs
            rel_RR(ii) = (RR(ii+1)/RR(ii)); % compare the lenght from the current interval to the subsequent one
        end
    end
    % calculate the score of the current signal, a lower score means a beter suitability for analysis
    % if not atleast 15 annotations have been generated or if the MPI is more than 2 seconds automatically assign a score of 100
    if exist('rel_RR','var') == 0 || numel(RR) < 15 || MPI > Fs * 2
        mean_idx(i) = 100; % assign score of 100
    else
        mean_idx(i) = std(rel_RR); % calculate score by determining the standard devation of the relative intervals
    end
    clear rel_RR
end
if sum(mean_idx) == 100*numel(mean_idx)
    ann_bpa = []; % if all candidates generate a score of 100, don't return any annotations
else
    [~,idx] = min(mean_idx); % find candidate with best score
    BPS = signalBU(:,idx+1); % select signal with best score
    % filter this signal
   fc = [1 5]*2*(1/Fs);                                                   % set cut-off frequencies
   [B,A] = butter(3,fc);                                        % construct bandpass filters
   BPS = filtfilt(B,A,BPS');                                    % filter bp-signal
    ann_bpa = ABP(BPS,0.375*FsBU,0.05*FsBU);			         % regenerate annotations using original signal and orignal sampling frequency
end
end
function [ann_ecg,delay] = check_PM(recordName,ann_ecg,ann_bpa,Fs) % check for the presence of a Pacemaker
% ann_ecg = annotations based on peaks in ecg
% ann_bpa = annotations based on peaks in bp-signal
% Fs = sampling frequency
% annotations are corrected when the algorithms senses the precense of a
% PM, when PM is absent annotations are not altered.

delay = check_delay(ann_ecg,ann_bpa,Fs); % this function checks the avarage delay between the abp- and ecg-annotations
IEA = diff(ann_ecg); % intervals between ecg-annotations
IEA(IEA>2*Fs) = [];
[freq,power] = hist(IEA,100); % histogram of ecg-annotation-intervals
[~,idx] = max(freq); % find most frequent interval in ecg-annotations
p1_ecg = power(idx); freq(idx) = []; power(idx) = [];
[~,idx] = max(freq); % find second most frequent interval in ecg-annotations
p2_ecg = power(idx);
p_ecg = p1_ecg+p2_ecg; % sum of most two most frequent intervals
IBA = diff(ann_bpa); % intervals between ecg-annotations
IBA(IBA>2*Fs) = [];
[freq,power] = hist(IBA,100); %
[~,idx] = max(freq);
p_bpa = power(idx); % most frequent interval in bpa-annotations


if abs(p_bpa-p_ecg) < 0.05*Fs % both pulses are annotated
    
    gqrs(recordName,[],[],[],10,'qrs',[])
    ann_ecg1 = rdann(recordName,'qrs');
    i = numel(ann_ecg1);
    while i > 1
        val = ann_ecg1(i);
        tmp = abs(ann_ecg-val);
        [~,idx] = min(tmp);
        ann_ecg(idx) = [];
        i = i - 1;
    end
    delay = check_delay(ann_ecg,ann_bpa,Fs); % re-calculate delay
elseif delay > 0.4*Fs % only the pm-pulse is annotated
    
    if numel(ann_ecg) > 1.2*numel(ann_bpa)
        gqrs(recordName,[],[],[],2,'qrs',[])
        ann_ecg = rdann(recordName,'qrs');
    else
        gqrs(recordName,[],[],[],.5,'qrs',[])
        ann_ecg1 = rdann(recordName,'qrs');
        if numel(ann_ecg1) < 2.2*numel(ann_ecg) && numel(ann_ecg1) > 1.8*numel(ann_ecg)
            i = numel(ann_ecg);
            while i > 1
                val = ann_ecg(i);
                tmp = abs(ann_ecg1-val);
                [~,idx] = min(tmp);
                ann_ecg1(idx) = [];
                i = i - 1;
            end
            ann_ecg = ann_ecg1;
            ann_ecg = ann_ecg';
        end
    end
    delay = check_delay(ann_ecg,ann_bpa,Fs); % re-calculate delay
end
end
function [ann_bpa] = ABP(BPS,MAI,MAW)
% ann_bpa = BP annotations, BPS = BP-signal, Fs = Sampling frequency, MAI = maximal allowed interval and MAW = maximal allowed width
D_s_filt = diff(BPS);                                                % derative of filtered signal
D_s_filt(D_s_filt < 0) = 0;                                             % set all signal with non-positve slope to zero
[freq,power] = hist(D_s_filt,100);                                      % determine histogram of modified signal
CS = cumsum(freq);
K = abs(CS-0.9*(CS(end)));
[~,bm] = min(K);
th = 0.3*power(bm);
[fff,~] = hist(BPS,100);
[~,idx] = max(fff);
ann_bpa = [];
if fff(idx)/sum(fff) < 0.50 && power(end) > 1 % determine percieved peak magnitude
    [ann_bpa] = FP(D_s_filt,round(MAI));      % identify prominent peaks as bp-pulses (min peak distance is at least .5 seconds)
    ii = numel(ann_bpa);                                                    % num-loop iterations
    for i = 1:length(ann_bpa)
        if D_s_filt(ann_bpa(ii)) < th;                                      % remove peaks below threshold
            ann_bpa(ii) = [];
        end
        ii = ii - 1;
    end
    if MAW > 0
        i = length(ann_bpa);
        while i>0 && numel(ann_bpa) > 1
            if i == 1
                A = D_s_filt(1:ann_bpa(i+1));
                idex = ann_bpa(i);
            elseif i == length(ann_bpa)
                A = D_s_filt(ann_bpa(i-1):end);
                idex = 1 + ann_bpa(i)-ann_bpa(i-1);
            else
                A = D_s_filt(ann_bpa(i-1):ann_bpa(i+1));
                idex = 1 + ann_bpa(i)-ann_bpa(i-1);
            end
            A = find(A==0);
            tmp = abs(A-idex);
            [~,idx] = min(tmp);
            X = A(idx:end);
            if X(1)<idex && numel(X)>1;
                X = X(2);
            else
                X = X(1);
            end
            if isempty(X) == 0
                Y = A(1:idx);
                if Y(end) > idex && numel(Y)>1
                    Y = Y(end-1);
                else
                    Y = Y(end);
                end
                if isempty(Y) == 0
                    width = X-Y;
                    if width < MAW
                        ann_bpa(i) = [];
                    end
                end
                i = i - 1;
            else
                i = 0;
            end
        end
    end
end
end
function [locs] = FP(data,mpd)
% find peak locations in data, with minimal peak distance in samples (mpd)

if iscolumn(data) == 0
    data = data';
end
d1 = diff(data);
d2 = diff(d1);
locs = find(d1.*[d1(2:end);0]<0);
locs(d2(locs)>=0) = [];
locs = locs + 1;
[a,b] = min(diff(locs));
while a < mpd;
    if data(locs(b)) > data(locs(b+1))
        locs(b+1) = [];
    else
        locs(b) = [];
    end
    [a,b] = min(diff(locs));
end
locs = locs';
end
function delay = check_delay(ann_ecg,ann_bpa,Fs) % find the avarage delay between the ECG-annotations and the ABP-annotations
ary = zeros(1,length(ann_ecg));

for yy = 1:length(ann_ecg)
    a = find(ann_bpa>ann_ecg(yy));
    if isempty(a) == 0
        ary(yy) = ann_bpa(a(1))-ann_ecg(yy);
    end
end

ary(ary>Fs) = [];
[fr,pw] = hist(ary,100);
[~,idx] = max(fr);
delay = round(pw(idx)); % calculate delay using ecg- and bpa-annotations
end
function [index] = HLC(peak_location,signal,qrs_width) % horizontal line check
% if peak is noisy, ind = 1

excerpt = signal(peak_location - round(0.5*qrs_width):peak_location + round(0.5*qrs_width));

Max = max(excerpt);
Min = min(excerpt);
diff = Max-Min;
num_lines = 20;
step = diff/(num_lines-1);
index = 0; i = 0;

while index == 0 && i < (num_lines+1)
    i = i+1;
    intsct = 0;
    level = Max - ((i-1)*step);
    for ii = 1:(length(excerpt)-1)
        if excerpt(ii) > level && excerpt(ii+1) < level
            intsct = intsct + 1;
        elseif excerpt(ii) < level && excerpt(ii+1) > level
            intsct = intsct + 1;
        end
    end
    if intsct > 6
        index = 1;
        %'noisy peak'
    end
end
end